/*--------------------------------------------------------------------
 *    The GMT-system: 	@(#)grd2cpt.c	2.47  02/22/00
 *
 *	Copyright (c) 1991-2000 by P. Wessel and W. H. F. Smith
 *	See COPYING file for copying and redistribution conditions.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; version 2 of the License.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	Contact info: www.soest.hawaii.edu/gmt
 *--------------------------------------------------------------------*/
/*
 * grd2tga read a 2d binary gridded grdfile and convert to tga image file or 
 * convert to tga height-field for POV 
 *
 *	  Persistence of Vision (POV) raytracer can use a height-field
 *	  defined in a Targa (.TGA) image file format where the RGB pixel values
 *	  are 24 bits (3 bytes). A 16 bit unsigned integer height-field value
 *	  is assigned as follows:
 *
 *		   Red     high byte
 *		   Green   low byte
 *		   Blue    empty
 * 
 * A true tga image file is a simple 24 bit image file format recognized by most
 * of image editing programs. The color schema used is the one described in
 * grd2cpt. 
 *
 * Author:	Joaquim Luis 
 *		Based on grd2cpt and dem2pov
 * Date:	26-MAY-2000
 * Revised:	26-MAY-2000, for GMT 3.3.1
 * 
 */

#include "gmt.h"

struct CDF_CPT {
	double	z;	/* Data value  */
	double	f;	/* Cumulative distribution function f(z)  */
} *cdf_cpt = NULL;

struct RGB_SLICE {
	double 	z_low, z_high;
	int 	rgb_low[3], rgb_high[3];
} *slice = NULL;

void  write_tga_header(FILE *, int width, int height);
void GMT_sample_a_cpt (double z[], int nz, BOOLEAN continuous, BOOLEAN reverse, struct RGB_SLICE *slice);
void write_true_tga (FILE *, int nxy, int ncdf, BOOLEAN continuous, float *zdata, struct RGB_SLICE *slice);
void write_pov_tga (FILE *, float v_scale, float *zdata, struct GRD_HEADER *grd);

main (int argc, char **argv)
{

	int i, j, nxy, nfound, ngood, ncdf, nofound = 0, r, g, b, pad[4];
	
	BOOLEAN error = FALSE, set_limits = FALSE, set_z_vals = FALSE, reverse = FALSE, ok = FALSE, continuous = FALSE, pov = FALSE;
	
	double *z, min_limit, max_limit, z_start, z_stop, z_inc, mean, sd, f;
	
	char *table = CNULL, *grdfile = CNULL, CPT_lis[BUFSIZ], CPT_file[BUFSIZ], format[BUFSIZ];
	char *outfile = CNULL;
	struct GRD_HEADER grd;

	float	*zdata, v_scale = 1.0;

	FILE	*fpc = NULL, *fp = NULL;
	pad[3] = pad[2] = pad[1] = pad[0] = 0;
		
	argc = GMT_begin (argc, argv);

	/* Get list of available color tables in GMT /lib */

	sprintf (CPT_lis, "%s%clib%cGMT_CPT.lis\0", GMTHOME, DIR_DELIM, DIR_DELIM);

	if ((fpc = fopen (CPT_lis, "r")) == NULL) {
		fprintf (stderr, "%s: ERROR: Cannot open file %s\n", GMT_program, CPT_lis);
		exit (EXIT_FAILURE);
	}

	for (i = 1; !error && i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
			
				/* Common parameters */
			
				case 'V':
				case '\0':
					error += GMT_get_common_args (argv[i], 0, 0, 0, 0);
					break;

				/* Supplemental parameters */
				
				case 'C':	/* Get cpt table */
					table = &argv[i][2];
					break;
				case 'G':
					outfile = &argv[i][2];
					break;
				case 'I':
					reverse = TRUE;
					break;

				case 'L':
					if (sscanf(&argv[i][2], "%lf/%lf", &min_limit, &max_limit) != 2) {
						fprintf(stderr,"%s: GMT SYNTAX ERROR -L option:  Cannot decode limits\n", GMT_program);
						error++;
					}
					else {
						if (min_limit >= max_limit) {
							fprintf(stderr,"%s: GMT SYNTAX ERROR -L option:  min_limit must be less than max_limit.\n", GMT_program);
							error++;
						}
					}
					if (!error) set_limits = TRUE;
					break;

				case 'S':
					if (sscanf(&argv[i][2], "%lf/%lf/%lf", &z_start, &z_stop, &z_inc) != 3) {
						fprintf(stderr,"%s: GMT SYNTAX ERROR -S option:  Cannot decode values\n", GMT_program);
						error++;
					}
					else {
						if (z_stop <= z_start || z_inc <= 0.0) {
							fprintf(stderr,"%s: GMT SYNTAX ERROR -S option:  Bad arguments\n", GMT_program);
							error++;
						}
					}
					if (!error) set_z_vals = TRUE;
					break;
				case 'P':
					pov = TRUE;
					break;
				case 'T':
					sscanf (&argv[i][2], "%f", &v_scale);
					break;
				case 'Z':
					continuous = TRUE;
					break;

				default:
					error = TRUE;
					GMT_default_error (argv[i][1]);
					break;
			}
		}
		else {
			grdfile = argv[i];
			if (GMT_read_grd_info (grdfile, &grd)) {
				fprintf (stderr, "%s: Error opening file %s\n", GMT_program, grdfile);
				error++;
			}
		}
	}
	
	if (argc == 1 || GMT_quick) {
		fprintf (stderr, " grd2tga %s - read a grdfile and convert to TGA image file or a TGA\n height-field file for POV-ray \n", GMT_VERSION);

		fprintf (stderr," In case that the user choses to create a true TGA image file (default) the\n options are the same as in grd2cpt. Note that if the grdfile has NaNs\n the NaN color can only be modified in .gmtdefaults.\n\n");
		fprintf (stderr, "usage: grd2tga <grdfile> -G<outfile> [-C<table>] [-I] [-L<min_limit>/<max_limit>]\n");
		fprintf (stderr, "	[-S<z_start>/<z_stop>/<z_inc>] [-P] [-T<scale_z_axis>] [-V] [-Z]\n\n");
		
		if (GMT_quick) exit (EXIT_FAILURE);
		
		fprintf (stderr, "\t<grdfile> is the name of the 2-D binary data set\n");
		fprintf (stderr, "\t-G<outfile> is the name of the TGA image file\n\n");
		fprintf (stderr, "\n\tOPTIONS:\n");
		fprintf(stderr,"\t-C  Specify a colortable [Default is rainbow]:\n");
		fprintf(stderr,"\t   [Original z-range is given in brackets]\n");
		fprintf(stderr,"\t   ---------------------------------\n");
		while (fgets (format, BUFSIZ, fpc)) if (!(format[0] == '#' || format[0] == 0)) fprintf (stderr, "\t   %s", format);
		fclose (fpc);
		fprintf(stderr,"\t   ---------------------------------\n");
		fprintf(stderr,"\t-I reverses the sense of the color table\n");
		fprintf (stderr, "	-L Limit the range of the data [Default uses actual min,max of data].\n");
		fprintf (stderr, "	-S Sample points should Step from z_start to z_stop by z_inc [Default guesses some values].\n");
		GMT_explain_option ('V');
		fprintf(stderr,"\t-Z will create a continuous color palette.\n");
		fprintf(stderr,"\t   [Default is discontinuous, i.e., constant color intervals]\n");
		fprintf(stderr,"\t-P will create a POV-ray tga file [Default creates a true tga].\n");
		fprintf (stderr, "\t-T scale z axis by this factor. Vertical must be in meters positive up. If your grid has z positive down use -1. Also useful if the grid is in km or feet [Default = 1]\n");
		exit (EXIT_FAILURE);
	}
	
	fclose (fpc);

	if (!outfile) {
		fprintf (stderr, "%s: GMT SYNTAX ERROR option -G:  Must specify output file\n", GMT_program);
		error++;
	}

	/* Open the specified master color table */
	/* First try current directory */

	if (table) {
		if (strstr (table, ".cpt"))
			strcpy (CPT_file, table);
		else
			sprintf (CPT_file, "%s.cpt\0", table);

		ok = !access (CPT_file, R_OK);
		if (ok && gmtdefs.verbose) fprintf (stderr, "%s: Reading %s in current directory\n", GMT_program, CPT_file);
	}

	if (!ok)	{	/* No table in current dir, try /lib */
		if (table)
			sprintf (CPT_file, "%s%clib%cGMT_%s.cpt\0", GMTHOME, DIR_DELIM, DIR_DELIM, table);
		else	/* Default to rainbow colors */
			sprintf (CPT_file, "%s%clib%cGMT_rainbow.cpt\0", GMTHOME, DIR_DELIM, DIR_DELIM);

		if (access (CPT_file, R_OK)) {
			fprintf (stderr, "%s: ERROR: Cannot find colortable %s\n", GMT_program, CPT_file);
			error++;
		}
		else if (gmtdefs.verbose)
			fprintf (stderr, "%s: Reading %s\n", GMT_program, CPT_file);
	}

	if (error) exit (EXIT_FAILURE);

	GMT_put_history (argc, argv);	/* Update .gmtcommands */

	GMT_read_cpt (CPT_file);

	nxy = grd.nx * grd.ny;
	zdata = (float *) GMT_memory (VNULL, (size_t)nxy, sizeof (float), GMT_program);

	if (GMT_read_grd (grdfile, &grd, zdata, 0.0, 0.0, 0.0, 0.0, pad, FALSE) ) {
		fprintf(stderr,"%s:  Error reading grdfile.\n", GMT_program);
		free ((void *)zdata);
		exit (EXIT_FAILURE);
	}

	if ((fp = fopen(outfile, "wb")) == NULL) {
		fprintf(stderr, "%s: Error can't open output file %s\n", GMT_program, outfile);
		exit(1);
	}

	if (pov) {
		/* write TGA image file header */
		write_tga_header (fp, grd.nx, grd.ny);
	
		write_pov_tga (fp, v_scale, zdata, &grd);
		free ((void *)zdata);
		exit (1);
	}
	/* Loop over the file and find NaNs.  If set limits, may create more NaNs  */
	nfound = 0;
	mean = 0.0;
	sd = 0.0;
	if (set_limits) {
		/* Loop over the grdfile, and set anything outside the limiting values to NaN.  */

		grd.z_min = min_limit;
		grd.z_max = max_limit;
		for (i = 0; i < nxy; i++) {
			if (GMT_is_fnan (zdata[i]))
				nfound++;
			else {
				if (zdata[i] < min_limit || zdata[i] > max_limit) {
					nfound++;
					zdata[i] = GMT_f_NaN;
				}
				else {
					mean += zdata[i];
					sd += zdata[i] * zdata[i];
				}
			}
		}
	}
	else {
		min_limit = grd.z_max;	/* This is just to double check grd.z_min, grd.z_max  */
		max_limit = grd.z_min;
		for (i = 0; i < nxy; i++) {
			if (GMT_is_fnan (zdata[i]))
				nfound++;
			else {
				if (zdata[i] < min_limit) min_limit = zdata[i];
				if (zdata[i] > max_limit) max_limit = zdata[i];
				mean += zdata[i];
				sd += zdata[i] * zdata[i];
			}
		}
		grd.z_min = min_limit;
		grd.z_max = max_limit;
	}
	ngood = nxy - nfound;	/* This is the number of non-NaN points for the cdf function  */
	mean /= ngood;
	sd /= ngood;
	sd = sqrt(sd - mean*mean);
	if (gmtdefs.verbose) {
		sprintf(format,"%%s:  Mean and S.D. of data are %s %s\n\0", gmtdefs.d_format, gmtdefs.d_format);
		fprintf(stderr, format, GMT_program, mean, sd);
	}

	/* Now the zdata are ready.  Decide how to make steps in z.  */
	if (set_z_vals) {
		ncdf =  (grd.z_min < z_start) ? 1 : 0;
		ncdf += (int)floor((z_stop - z_start)/z_inc) + 1;
		if (grd.z_max > z_stop) ncdf++;
		cdf_cpt = (struct CDF_CPT *)GMT_memory (VNULL, (size_t)ncdf, sizeof(struct CDF_CPT), GMT_program);
		if (grd.z_min < z_start) {
			cdf_cpt[0].z = grd.z_min;
			cdf_cpt[1].z = z_start;
			i = 2;
		}
		else {
			cdf_cpt[0].z = z_start;
			i = 1;
		}
		j = (grd.z_max > z_stop) ? ncdf - 1 : ncdf;
		while (i < j) {
			cdf_cpt[i].z = cdf_cpt[i-1].z + z_inc;
			i++;
		}
		if (j == ncdf-1) cdf_cpt[j].z = grd.z_max;
	}
	else {
		/* This is completely ad-hoc.  It chooses z based on steps
			of 0.1 for a Gaussian CDF:  */
		ncdf = 11;
		cdf_cpt = (struct CDF_CPT *)GMT_memory (VNULL, (size_t)ncdf, sizeof(struct CDF_CPT), GMT_program);
                /* Stupid bug fix here:  
                        If (mean-1.28155*sd <= grd.z_min || mean+1.28155*sd >= grd.z_max) then
                          reset mean and sd so they fit inside available range:  */
                          
		if ((mean - 1.28155*sd) <= grd.z_min || (mean + 1.28155*sd) >= grd.z_max) {
			mean = 0.5 * (grd.z_min + grd.z_max);
			sd = (grd.z_max - mean) / 1.5;
			if (sd <= 0.0) {
				fprintf (stderr, "%s:  ERROR.  Min and Max data values are equal.\n", GMT_program);
				exit (EXIT_FAILURE);
			}
		}	/* End of stupid bug fix  */
		
		cdf_cpt[0].z = grd.z_min;
		cdf_cpt[1].z = mean - 1.28155 * sd;
		cdf_cpt[2].z = mean - 0.84162 * sd;
		cdf_cpt[3].z = mean - 0.52440 * sd;
		cdf_cpt[4].z = mean - 0.25335 * sd;
		cdf_cpt[5].z = mean;
		cdf_cpt[6].z = mean + 0.25335 * sd;
		cdf_cpt[7].z = mean + 0.52440 * sd;
		cdf_cpt[8].z = mean + 0.84162 * sd;
		cdf_cpt[9].z = mean + 1.28155 * sd;
		cdf_cpt[10].z = grd.z_max;
	}

	/* Get here when we are ready to go.  cdf_cpt[].z contains the sample points.  */

	if (gmtdefs.verbose) sprintf (format, "%%s: z = %s and CDF(z) = %s\n\0", gmtdefs.d_format, gmtdefs.d_format);
	for (j = 0; j < ncdf; j++) {
		if (cdf_cpt[j].z <= grd.z_min)
			cdf_cpt[j].f = 0.0;
		else if (cdf_cpt[j].z >= grd.z_max)
			cdf_cpt[j].f = 1.0;
		else {
			nfound = 0;
			for (i = 0; i < nxy; i++) {
				if (!GMT_is_fnan (zdata[i]) && zdata[i] <= cdf_cpt[j].z) nfound++;
			}
			cdf_cpt[j].f = (double)(nfound-1)/(double)(ngood-1);
		}
		if (gmtdefs.verbose) fprintf (stderr, format, GMT_program, cdf_cpt[j].z, cdf_cpt[j].f);
	}

	/* Now the cdf function has been found.  We now resample the chosen cptfile  */

	z = (double *) GMT_memory (VNULL, (size_t)ncdf, sizeof (double), GMT_program);
	for (i = 0; i < ncdf; i++) z[i] = cdf_cpt[i].z;

	slice = (struct RGB_SLICE *) GMT_memory (VNULL, (size_t)ncdf, sizeof (struct RGB_SLICE), GMT_program);

	GMT_sample_a_cpt (z, ncdf, continuous, reverse, slice);

	/* write TGA image file header */
	write_tga_header (fp, grd.nx, grd.ny);
	
	write_true_tga (fp, nxy, ncdf, continuous, zdata, slice);

	free ((void *)cdf_cpt);
	free ((void *)slice);
	free ((void *)zdata);
	free ((void *)z);
	
	GMT_end (argc, argv);
}

void GMT_sample_a_cpt (double z[], int nz, BOOLEAN continuous, BOOLEAN reverse, struct RGB_SLICE *slice)
{
	/* Resamples the current cpt table based on new z-array.
	 * Old cpt is normalized to 0-1 range and scaled to fit new z range.
	 * New cpt may be continuous and/or reversed.
	 * We write the new cpt table to stdout */

	int i, j, k, nx, upper, lower, rgb_low[3], rgb_high[3];
	double *x, a, b, f, x_inc;
	char format[BUFSIZ];
	struct GMT_LUT *lut;

	lut = (struct GMT_LUT *) GMT_memory (VNULL, (size_t)GMT_n_colors, sizeof (struct GMT_LUT), GMT_program);
	
	/* First normalize old cpt file so z-range is 0-1 */

	b = 1.0 / (GMT_lut[GMT_n_colors-1].z_high - GMT_lut[0].z_low);
	a = -GMT_lut[0].z_low * b;

	for (i = 0; i < GMT_n_colors; i++) {	/* Copy/normalize cpt file and reverse if needed */
		lut[i].z_low = a + b * GMT_lut[i].z_low;
		lut[i].z_high = a + b * GMT_lut[i].z_high;
		if (reverse) {
			j = GMT_n_colors - i - 1;
			memcpy ((void *)lut[i].rgb_high, (void *)GMT_lut[j].rgb_low,  (size_t)(3 * sizeof (int)));
			memcpy ((void *)lut[i].rgb_low,  (void *)GMT_lut[j].rgb_high, (size_t)(3 * sizeof (int)));
		}
		else {
			j = i;
			memcpy ((void *)lut[i].rgb_high, (void *)GMT_lut[j].rgb_high, (size_t)(3 * sizeof (int)));
			memcpy ((void *)lut[i].rgb_low,  (void *)GMT_lut[j].rgb_low,  (size_t)(3 * sizeof (int)));
		}
	}
	
	/* Then set up normalized output locations x */

	nx = (continuous) ? nz : nz - 1;
	x = (double *) GMT_memory (VNULL, (size_t)nz, sizeof(double), GMT_program);

	if (nx == 1) {
		x[0] = 0.5;
	}
	else {
		x_inc = 1.0 / (nx - 1);
		for (i = 0; i < nz; i++) x[i] = i * x_inc;	/* Normalized z values 0-1 */
	}


	for (i = 0; i < nz-1; i++) {
		lower = i;
		upper = i + 1;
		
		/* Interpolate lower color */

		for (j = 0; j < GMT_n_colors && x[lower] >= lut[j].z_high; j++);
		if (j == GMT_n_colors) j--;

		f = 1.0 / (lut[j].z_high - lut[j].z_low);
		
		for (k = 0; k < 3; k++) slice[i].rgb_low[k] = lut[j].rgb_low[k] + irint ((lut[j].rgb_high[k] - lut[j].rgb_low[k]) * f * (x[lower] - lut[j].z_low));

		if (continuous) {	/* Interpolate upper color */

			while (j < GMT_n_colors && x[upper] > lut[j].z_high) j++;

			f = 1.0 / (lut[j].z_high - lut[j].z_low);
			
			for (k = 0; k < 3; k++) slice[i].rgb_high[k] = lut[j].rgb_low[k] + irint ((lut[j].rgb_high[k] - lut[j].rgb_low[k]) * f * (x[upper] - lut[j].z_low));
		}
		else	/* Just copy lower */
			for (k = 0; k < 3; k++) slice[i].rgb_high[k] = slice[i].rgb_low[k];

		slice[i].z_low = z[i];		slice[i].z_high = z[i+1];
	}

	free ((void *)x);
	free ((void *)lut);
}
/*****************************************************************************/
void write_tga_header(FILE *outputFile, int width, int height)
{
   int i;

   /* Write TGA image header */
   for (i = 0; i < 10; i++)      /* 00, 00, 02, then 7 00's... */
      if (i == 2)
	 putc((short)i, outputFile);
      else
	 putc((short)0, outputFile);

   putc(0, outputFile); /* y origin set to "First_Line" */
   putc(0, outputFile);

   putc((short)(width % 256), outputFile);  /* write width and height */
   putc((short)(width / 256), outputFile);
   putc((short)(height % 256), outputFile);
   putc((short)(height / 256), outputFile);
   putc(24, outputFile);  /* 24 bits/pixel (16 million colors!) */
   putc(32, outputFile);  /* Bitmask, pertinent bit: top-down raster */
}

/******************************************************************************/

void write_true_tga (FILE *fp, int nxy, int ncdf, BOOLEAN continuous, float *zdata, struct RGB_SLICE *slice) { 
	int i, j, r, g, b, nfound = 0;
	double f;

	for (i = 0; i < nxy; i++) {
		for (j = 0; j < ncdf-1; j++) {
			if (zdata[i] >= slice[j].z_low && zdata[i] <= slice[j].z_high) {
				if (continuous) {
					f = (zdata[i] - slice[j].z_low) / (slice[j].z_high - slice[j].z_low);
					r = slice[j].rgb_low[2] + irint (f * (slice[j].rgb_high[2] - slice[j].rgb_low[2]));
					g = slice[j].rgb_low[1] + irint (f * (slice[j].rgb_high[1] - slice[j].rgb_low[1]));
					b = slice[j].rgb_low[0] + irint (f * (slice[j].rgb_high[0] - slice[j].rgb_low[0]));
					/* write pixel */
					putc((char)(r), fp);
					putc((char)(g), fp);
					putc((char)(b), fp);
				}
				else {
					/* write pixel */
					putc((char)(slice[j].rgb_low[2]), fp);
					putc((char)(slice[j].rgb_low[1]), fp);
					putc((char)(slice[j].rgb_low[0]), fp);
				}
				break;
			}
			else if (j == ncdf -2) {
				putc((char)(GMT_bfn.nan_rgb[2]), fp);
				putc((char)(GMT_bfn.nan_rgb[1]), fp);
				putc((char)(GMT_bfn.nan_rgb[0]), fp);
			}
		}
	}

}
/******************************************************************************/
void write_pov_tga (FILE *fp, float v_scale, float *zdata, struct GRD_HEADER *grd) { 
	int i, j, ij, nxy;
	float tmp, hfnorm = 1.0, tmin, tmax;

	nxy = grd->nx * grd->ny;
	tmin = 65535;	tmax = -65535;

	if (v_scale < 0) { /* z was positive down */
		for (i = 0; i < nxy; i++) zdata[i] = zdata[i] * v_scale + fabs(v_scale) * grd->z_max;
		hfnorm = 65535.0/(fabs(v_scale) * (grd->z_max - grd->z_min));
	}
	else {
		for (i = 0; i < nxy; i++) zdata[i] = zdata[i] * v_scale - v_scale * grd->z_min;
		hfnorm = 65535.0/(v_scale * (grd->z_max - grd->z_min));
	}

	for(i = 0; i < nxy; i++) {
		tmp = zdata[i]; 
		/* Normalize */
		tmp *= hfnorm;
		/*if (!GMT_is_fnan (tmp)) {
			tmin = MIN (tmin, tmp);
			tmax = MAX (tmax, tmp);
		}*/
		if(tmp < 0.0) tmp = 0.0;
		if(tmp > 65535.0) tmp = 65535.0;
		if (GMT_is_fnan (tmp)) tmp = 0.0;

		/* write pixel */
		putc((char)0, fp);/* Blue  empty     */
		putc((char)((int) tmp % 256), fp); /* Green low byte  */
		putc((char)((int) tmp / 256), fp); /* Red   high byte */
	}
	/*fprintf(stderr, "min= %f max= %f\n", tmin, tmax);*/
}
